//
// GameX - ImageX Class Code 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 


// DirectDraw
#ifndef DIRECTDRAW_VERSION
#define DIRECTDRAW_VERSION 0x0700 // specify version 7.0 of DirectDraw
//#include <ddraw.h>
#endif

// Direct3D
#ifndef DIRECT3D_VERSION
#define DIRECT3D_VERSION 0x0700 // specify version 7.0 of Direct3D -- the last version that's compatible with DirectDraw
#include <d3d.h>				// note: for newer versions of D3D we would need to 
//#include <d3dtypes.h>			//		 include different files, like d3d8.h
//#include <d3dx.h>
#endif


//#include <math.h>
//#include <string.h>
#include "gamex-image.h"
#include "gamex-win-dx.h"

// disable memory tracking of this file unless we're building debug (otherwise it would go into the release library)
#ifndef _DEBUG
#define CLEANMEM_DISABLE
#include "gamex-cleanmem.h"
#endif


ImageX::ImageX ()
{
	z_data = NULL;
	z_alpha = NULL;
	z_surface = NULL;
	z_ytable = NULL;
	z_copied = NULL;
	z_load_queued = false;
	z_copyable = false;
	for(int i = 0 ; i < 3 ; i++) z_internal_compatibility_img[i] = NULL;
	Reset();
	sprintf(z_filename, "Uninitialized ImageX");
}

ImageX::~ImageX ()
{
	Destroy();
}

void ImageX::Reset ()
{
	z_xres = 0;
	z_yres = 0;
	z_size = 0;
	z_options = IMG_UNDEF;
	DeleteData(); // delete the non-surface data
	z_usage = DRAW_UNDEF;
	z_status = IMG_NOTREADY;
	z_hassurface = 0;
	z_surface = NULL;
	z_device = NULL;
	ResetFilterRect();
	z_non_empty = false;
	z_load_queued = false;
	if(!z_copied)
		SAFE_DELETE_ARRAY(z_ytable);
	else
		z_ytable = NULL;

///	if(z_load_queued)
	{
		// unqueue ourself from being loaded
		if(z_GameX_exists)
		{
			std::vector<ImageX*>::iterator thisEntry = std::find(GameX.z_queued_image_loads.begin(), GameX.z_queued_image_loads.end(), this);
			if(thisEntry != GameX.z_queued_image_loads.end())
				GameX.z_queued_image_loads.erase(thisEntry);
		}
	}

	if(z_copied)
		z_UnregisterCopied();

	if(z_copyable)
		z_UnregisterCopyable();

}

void ImageX::z_UnregisterCopied ()
{
	assertNoLib(z_copied);

	z_copied = NULL;

	// remove this image from the copying list
	if(z_GameX_exists)
	{
		std::vector<ImageX*>::iterator thisEntry = std::find(GameX.z_copying_images.begin(), GameX.z_copying_images.end(), this);
		if(thisEntry != GameX.z_copying_images.end())
			GameX.z_copying_images.erase(thisEntry);
	}
}

bool ImageX::z_CopyIdentical ()
{
	std::vector<ImageX*>::iterator end = GameX.z_copyable_images_loaded.end();
	for(std::vector<ImageX*>::iterator iter = GameX.z_copyable_images_loaded.begin() ; iter != end ; iter++)
	{
		ImageX* item = *iter;
		if(item->z_copyable && !item->z_load_queued
		&& _stricmp(z_filename, item->z_filename) == 0
		&& _stricmp(z_load_alpha_filename, item->z_load_alpha_filename) == 0
		&& z_load_alpha_bool == item->z_load_alpha_bool)
		{
			assertNoLib(this != item);

			// become a shallow copy of the image that was loaded identically
			memcpy(this, item, sizeof(ImageX));

			z_copied = item;
			z_copyable = false;
			GameX.z_copying_images.push_back(this);

			return true;
		}
	}
	return false;
}

void ImageX::z_RegisterCopyable ()
{
	assertNoLib(!z_copied);
	assertNoLib(!z_copyable);
	if(!z_copyable)
	{
		z_copyable = true;
		GameX.z_copyable_images_loaded.push_back(this);
	}
}

void ImageX::z_UnregisterCopyable ()
{
	assertNoLib(z_copyable);
	if(z_copyable)
	{
		z_copyable = false;

		if(!z_GameX_exists)
			return;

#ifndef _DEBUG // skip if shutting down, unless debugging GameX
		if(GameX.z_GetStatus() != GAMEX_SHUTTINGDOWN)
#endif
		{
			// remove this image from the copyable list
			std::vector<ImageX*>::iterator thisEntry = std::find(GameX.z_copyable_images_loaded.begin(), GameX.z_copyable_images_loaded.end(), this);
			if(thisEntry != GameX.z_copyable_images_loaded.end())
				GameX.z_copyable_images_loaded.erase(thisEntry);

			// for each thing that copied this, remove its copied status and put it back in the loading queue
			for(std::vector<ImageX*>::iterator iter = GameX.z_copying_images.begin() ; iter != GameX.z_copying_images.end() ; )
			{
				ImageX* item = *iter;
				assertNoLib(item && item->z_copied);
				if(item->z_copied == this)
				{
					assertNoLib(item != this);
					iter = GameX.z_copying_images.erase(iter);
					GameX.z_queued_image_loads.push_back(item);
				}
				else
				{
					iter++;
				}
			}
		}
	}
}

// hacks
static bool forceLoad = false;
bool z_imageCopyAllowed = true;

// Loads an image from file into this ImageX
bool ImageX::Load (const char * filename, bool alpha)
{
	assertMsg2(z_GameX_exists && GameX.z_win_hwnd, "Trying to load the image \"%s\" before GameX has been initialized.", REMOVE_TRAILING_SEPARATOR(filename));
	if(!z_GameX_exists || !GameX.z_win_hwnd) return false;	

	strcpy(z_filename, filename);
	z_load_alpha_filename[0] = '\0';
	z_load_alpha_bool = alpha;

	if(GameX.z_win_init_done || z_copied || forceLoad)
	{
		if(z_copied)
			z_UnregisterCopied();
		if(z_imageCopyAllowed && z_CopyIdentical())
			return true;

		ImageExt ie;
		bool ok = ie.Load(filename,this,alpha);

		if(ok)
			z_RegisterCopyable();

		return ok;
	}
	else
	{
		// queue for loading later
		if(!z_load_queued)
		{
			z_load_queued = true;
			GameX.z_queued_image_loads.push_back(this);
		}
		return true;
	}
}

bool ImageX::Load (const char * filename, const char * alphafilename)
{
	assertMsg3(z_GameX_exists && GameX.z_win_hwnd, "Trying to load the image \"%s\" + \"%s\" before GameX has been initialized.", REMOVE_TRAILING_SEPARATOR(filename), REMOVE_TRAILING_SEPARATOR(alphafilename));
	if(!z_GameX_exists || !GameX.z_win_hwnd) return false;	

	strcpy(z_filename, filename);
	strcpy(z_load_alpha_filename, alphafilename);
	z_load_alpha_bool = false;

	if(GameX.z_win_init_done || z_copied || forceLoad)
	{
		if(z_copied)
			z_UnregisterCopied();
		if(z_imageCopyAllowed && z_CopyIdentical())
			return true;

		// load it without the alpha
		ImageExt ie;
		bool ok1 = filename ? ie.Load(filename,this,true) : false;
		if(!ok1)
		{
			// if the non-alpha filename is invalid, treat it as black
			bool ok3 = ie.Load(alphafilename,this,true);
			if(!ok3) return false;
			else ChangeBrightness(-255,0);
		}
		// now load a temp image as the alpha and copy it into this image's alpha bits
		bool ok2 = LoadIntoAlpha(alphafilename);

		if(ok1 && ok2)
			z_RegisterCopyable();

		return (ok1 && ok2);
	}
	else
	{
		// queue for loading later
		if(!z_load_queued)
		{
			z_load_queued = true;
			GameX.z_queued_image_loads.push_back(this);
		}
		return true;
	}
}

void ImageX::ForceLoad ()
{
	if(z_copyable)
	{
		assertNoLib(!z_copied);
		z_UnregisterCopyable();
	}

	if(z_copied)
	{
		Reset();
		z_copied = NULL;
		z_copyable = false;
		z_load_queued = true;
	}

	if(z_load_queued)
	{
		forceLoad = true;

		if(z_load_alpha_filename[0])
			Load(z_filename, z_load_alpha_filename);
		else
			Load(z_filename, z_load_alpha_bool);

		forceLoad = false;

		z_load_queued = false;
	}
}

bool ImageX::LoadIntoAlpha (const char * filename)
{
	GameX.z_SetStatus(GAMEX_CREATING_TEST_SURFACE); // prevent GameX from registering as user-created
	ImageX temp_img;
	ImageExt ie;
	bool ok = ie.Load(filename, &temp_img);
	if(ok) temp_img.CopyToAlpha(this);
	GameX.z_SetStatus(GAMEX_READY);
	return ok;
}

// Creates a blank image, to be drawn after you draw into it
void ImageX::Create (int xr, int yr, bool alpha)
{
	ImageUsage use = (alpha) ? (LOAD_ALPHA) : 0;
	z_CreateWithUse(xr,yr,use);
}

void ImageX::z_CreateWithUse (int xr, int yr, ImageUsage use)
{
	Destroy();
	Reset();

	strcpy(z_filename,"Created ImageX");

	if(GameX.z_win_log_debug)
		if(GameX.z_GetStatus() == GAMEX_READY)
			debug.Output("Created Image:          "); 

	switch (use)
	{
		case LOAD_ALPHA: Size (xr, yr, IMG_TRUECOLORX | IMG_MERGEALPHA); break;	
		default: Size (xr, yr, IMG_HIGHCOLOR); break;
	}
	UseWith(use);
	Fill(0,0,0,0);
	z_non_empty = false;
}

void ImageX::Destroy ()
{
	if(z_surface != NULL)
	{
		if(!z_copied)
		{
			if(z_GameX_exists && GameX.z_GetDD()) { // if we haven't already released all of DirectDraw
				GameX.z_MemoryRecordChange(-z_surf_width*z_surf_height*(z_surf_bpp/8));
				z_surface->Release(); // release the surface data
			}
		}
		z_surface = NULL;
	}
	for(int i = 0 ; i < 3 ; i++)
	if(z_internal_compatibility_img[i] != NULL)
	{
		if(!z_copied)
			z_internal_compatibility_img[i]->Destroy();
		z_internal_compatibility_img[i] = NULL;
	}
	if(z_device != NULL)
	{
		if(!z_copied)
		{
			if(z_GameX_exists && GameX.z_GetDD()) { // if we haven't already released all of DirectDraw
				z_device->Release();
			}
		}
		z_device = NULL;
	}
	Reset();
}

void ImageX::DeleteData ()
{
	if(!z_copied)
	{
		SAFE_DELETE_ARRAY(z_data);
		SAFE_DELETE_ARRAY(z_alpha);
	}
	else
	{
		z_data = NULL;
		z_alpha = NULL;
	}
}

bool ImageX::Save (const char * filename)
{
	ImageExt ie;
	return ie.Save(filename,this);
}

void ImageX::ConvertUsageTo (ImageUsage use) 
{
	if(use == z_usage) return;
	z_PreservedRecreateAs(z_xres, z_yres, use);
}

void ImageX::Resize (int new_xres, int new_yres)
{
	if(new_xres == z_xres && new_yres == z_yres) return;
	z_PreservedRecreateAs(new_xres, new_yres, z_usage);
}

void ImageX::z_PreservedRecreateAs (int new_xres, int new_yres, ImageUsage new_use)
{
	int prevStatus = GameX.z_GetStatus();
	GameX.z_SetStatus(GAMEX_CREATING_TEST_SURFACE); // (prevent GameX from registering as user-created)
	if(z_non_empty)
	{
		ImageX temp_img;
		temp_img.z_CreateWithUse(new_xres, new_yres, new_use);
		CopyTo(&temp_img);
		z_CreateWithUse(temp_img.z_xres, temp_img.z_yres, temp_img.z_usage);
		temp_img.CopyTo(this);
	}
	else
	{
		z_CreateWithUse(new_xres, new_yres, new_use);
		Fill(0,0,0);
	}
	GameX.z_SetStatus(prevStatus);
}

void ImageX::z_AllocateSurface ()
{
	bool textureSurface = !(z_usage & LOAD_NON3D);
	int alphaSurface = ((z_usage & LOAD_ALPHA) && (z_options & IMG_ALPHA) && !(z_usage & LOAD_MASKED)) ? 2 : 0;
	bool targetable = (z_usage & LOAD_TARGETABLE) ? true : false;
	z_surface = GameX.z_CreateImageSurface(z_xres, z_yres, textureSurface, alphaSurface, targetable);

	if(z_surface != NULL)
	{

		DDPIXELFORMAT ddpf;
		ZeroMemory (&ddpf, sizeof(DDPIXELFORMAT)) ;
		ddpf.dwSize = sizeof(ddpf);
		z_surface->GetPixelFormat(&ddpf); // get the pixel format DirectDraw made the surface in,
		z_surf_bpp = ddpf.dwRGBBitCount; // usually this is z_win_screenbpp or 32 for alpha, but it's not guaranteed

		if(targetable && textureSurface)
		{
			// this will probably get the best SOFTWARE 3D rendering we have,
			// since the hardware device is likely to be tied up by the back buffer,
			// This is not the way to enable fastest drawing to surfaces, but in DX7 it seems to be the easiest and most compatible way
			if(GameX.z_CreateDeviceForSurface(z_device, z_surface))
			{

				z_device->SetTextureStageState(0,D3DTSS_ADDRESSU,D3DTADDRESS_CLAMP); // we don't want texture wrapping
				z_device->SetTextureStageState(0,D3DTSS_ADDRESSV,D3DTADDRESS_CLAMP); // that doesn't work right on most image sizes anyway
				z_device->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_MODULATE);
				z_device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_LINEAR);
				z_device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTFG_LINEAR);

				z_device->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE); // we have no D3D light sources, for now
				z_device->SetRenderState(D3DRENDERSTATE_COLORVERTEX, TRUE); // we want to allow per-vertex coloring
				z_device->SetRenderState(D3DRENDERSTATE_ANTIALIAS, FALSE); // set antialiasing off				
				z_device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE); // turn on alpha blending
				z_device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE); // and dithering

			}
		}

		z_hassurface = (textureSurface) ? 2 : 1; // success: 1 for normal, 2 for 3D accelerated

		DDSURFACEDESC2 ddsd;	
		ZeroMemory (&ddsd, sizeof (ddsd)) ;
		ddsd.dwSize = sizeof (ddsd) ;
		z_surface->GetSurfaceDesc(&ddsd);
		z_surf_width = ddsd.dwWidth;
		z_surf_height = ddsd.dwHeight;

		if(z_surf_width < z_xres)
			z_xmult = (float)z_surf_width / (float)z_xres;
		if(z_surf_height < z_yres)
			z_ymult = (float)z_surf_height / (float)z_yres;

		int bpp = GetBytesPerPixel(z_options);
		if(alphaSurface == 2)
			bpp = 4; // may not have converted it yet (such as masked to 32-bit alpha)
		int adjusted_pitch = ddsd.lPitch / bpp;
		SAFE_DELETE_ARRAY(z_ytable);
		z_ytable = new int[max(z_yres,z_surf_height)+32];
		for(int y = 0; y <= max(z_yres,z_surf_height); y++)
			z_ytable[y] = (int)((float)y * z_ymult) * adjusted_pitch;

		/// TEMP FIXME what was this here for? Possibly to allow clipping with custom viewports, but I'm not sure yet.
/*
		RECT boundary[1] = {0,0,z_xres,z_yres};
		LPDIRECTDRAWCLIPPER clipper;
		LPRGNDATA rgn_data;
		if(FAILED(GameX.z_GetDD()->CreateClipper(0,&clipper,NULL)))
		{
			return;
		}
		int rectSize = sizeof(RECT);
		int rgnHeaderSize = sizeof(RGNDATAHEADER);
		rgn_data = (LPRGNDATA)malloc(rgnHeaderSize+rectSize);
		if(!rgn_data)
			return;
		memcpy(rgn_data->Buffer, boundary, rectSize);
		rgn_data->rdh.dwSize = rgnHeaderSize;
		rgn_data->rdh.nRgnSize = rectSize;
		rgn_data->rdh.nCount = 1;
		rgn_data->rdh.iType = RDH_RECTANGLES;
		rgn_data->rdh.rcBound = boundary[0];
		if(FAILED(clipper->SetClipList(rgn_data,0)) || FAILED(z_surface->SetClipper(clipper)))
		{
		 	free(rgn_data);
			return;
		}
		free(rgn_data);*/
	}
	else
	{
		z_hassurface = 0; // failure
	}
}

void ImageX::z_DataToSurface ()
{
	// copy the pixels onto the surface, so that the surface can be blitted for us later
	if(z_hassurface == 0 || z_data == NULL)
	{
		return;
	}

	DDSURFACEDESC2 ddsd;	
	ZeroMemory (&ddsd, sizeof (ddsd)) ;
	ddsd.dwSize = sizeof (ddsd) ;

 	HRESULT z_status = z_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL); 	// lock drawing surface
	if(z_status != DD_OK)
	{
		GameX.z_DirectDrawError(z_status);
		return;
	}

	int xres = z_xres;
	int yres = z_yres;

	if(z_xmult != 1.0f || z_ymult != 1.0f)
	{

		// Scale the image's pixel data to reduce image resolution:

		int bytes_per_pix = GetBytesPerPixel(z_options);
		int bytes_per_alp = (z_alpha != NULL) ? 1 : 0;

		unsigned char *pData  = (unsigned char *) z_data; // pointer to data write pixel
		unsigned char *pDataO  = (unsigned char *) z_data; // pointer to data read pixel
		unsigned char *pAlpha = (unsigned char *) z_alpha; // pointer to alpha write pixel
		unsigned char *pAlphaO = (unsigned char *) z_alpha; // pointer to alpha read pixel

		const int adj_pix_pitch = z_xres * bytes_per_pix;
		const int adj_alp_pitch = z_xres;

		float yConvCounter = 0.0f;
		for(int sy = 0, dy = 0 ; sy < z_yres ; sy++)
		{
			float xConvCounter = 1.0f;
			if(yConvCounter >= 1.0f)
			{
				yConvCounter -= 1.0f;
				pData += adj_pix_pitch;
				pAlpha += adj_alp_pitch;
				dy++;
			}
			for(int sx = 0, dx = 0 ; sx < z_xres ; sx++)
			{
				if(xConvCounter >= 1.0f)
				{
					xConvCounter -= 1.0f;
					//copy
					for(int i = 0 ; i < bytes_per_pix ; i++)
						pData[dx*bytes_per_pix+i] = pDataO[sx*bytes_per_pix+i];
					for(int j = 0 ; j < bytes_per_alp ; j++)
						pAlpha[dx*bytes_per_alp+j] = pAlphaO[sx*bytes_per_alp+j];
					dx++;
				}
				xConvCounter += z_xmult;
			}
			yConvCounter += z_ymult;
			pDataO += adj_pix_pitch;
			pAlphaO += adj_alp_pitch;
		}

		xres = (int)((float)xres * z_xmult);
		yres = (int)((float)yres * z_ymult);
	}

	if(z_options & IMG_INDEXED)
	{
		GameX.z_SafeMessageBox("Error: Unsupported transfer operation (8 bit image to surface).", "GameX Error",MB_OK|MB_ICONSTOP) ;
		// not supported yet
	}
	else
	if(z_options & IMG_HIGHCOLOR)
	{
		if(z_options & IMG_ALPHA)
		{
			// if this image has a separate alpha channel,
			// combine the alpha and color data into a 32 bit surface texture for D3D to draw
			// (D3D's conversion from 32 bit to 16 bit on each draw can't be helped)
			unsigned short *pDataOld;
			unsigned char *pAlphaOld;
			unsigned long *pDataNew;
			pDataOld = (unsigned short *) z_data; 
			pAlphaOld = (unsigned char *) z_alpha; 
			pDataNew = (unsigned long *) ddsd.lpSurface;
			const int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			register int offset = 0;
			const int rMask = GameX.z_win_maskr;
			const int gMask = GameX.z_win_maskg;
			const int bMask = GameX.z_win_maskb;
			const int rShift = GameX.z_win_shiftr;
			const int gShift = GameX.z_win_shiftg;
			const bool extraX = z_surf_width > xres;
			const bool extraY = z_surf_height > yres;

			for(int y = 0 ; y < z_yres ; y++)
			{
				offset = y * adjusted_pitch;
				const bool yAtTop = y == 0 && extraY;
				const bool yAtBot = y == yres-1 && extraY;
				const bool yAtEdge = yAtTop || yAtBot;
				for(int x = 0 ; x < z_xres ; x++)
				{
					if(x < xres && y < yres)
					{
						const int clr = *pDataOld;
						const int newColor = RGBA_MAKE((clr & rMask) >> rShift << 3,
													   (clr & gMask) >> gShift << 2,
													   (clr & bMask) /*>> bShift*/ << 3,
													   z_alpha ? *pAlphaOld : 0);
						*(pDataNew + offset) = newColor;

						// clamp by one pixel to prevent oddities at scaled edges
						if(extraX)
						{
							if(x == xres-1)
								*(pDataNew + offset + 1) = newColor;
							else if(x == 0)
								*(pDataNew + offset + z_surf_width-1) = newColor;
						}
						if(yAtEdge)
						{
							if(yAtBot)
							{
								*(pDataNew + offset + adjusted_pitch) = newColor;
								if(x == xres-1 && extraX)
									*(pDataNew + offset + adjusted_pitch + 1) = newColor;
							}
							if(yAtTop)
								*(pDataNew + offset + adjusted_pitch*(z_surf_height-1)) = newColor;
						}

						offset++;
					}
					pDataOld++;
					pAlphaOld++;
				}
			}

		}
		else
		{
			unsigned short * buf_data = (unsigned short *) ddsd.lpSurface;
			const int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
		//	const int adjusted_pitch = ddsd.lPitch >> (GetBytesPerPixel(z_options)-1); // pitch adjusted for bit depth (>> 1 for 16 bit, etc.)
			const int adjusted_pitch = ddsd.lPitch / 2; // /2 for 16 bits
			unsigned short *pData = (unsigned short*) z_data;

			register int offset = 0;

			for(int y = 0 ; y < z_yres ; y++)
			{
				for(int x = 0 ; x < z_xres ; x++)
				{
					if(x < xres && y < yres)
					{
						const int newColor = *pData;
						*(buf_data + offset) = newColor;

						// clamp by one pixel to prevent oddities at scaled edges
						if(x == xres-1 && z_surf_width > xres)
							*(buf_data + offset + 1) = newColor;
						if(y == yres-1 && z_surf_height > yres)
						{
							*(buf_data + offset + adjusted_pitch) = newColor;
							if(x == xres-1 && z_surf_width > xres)
								*(buf_data + offset + adjusted_pitch + 1) = newColor;
						}
						if(x == 0 && z_surf_width > xres)
							*(buf_data + offset + z_surf_width-1) = newColor;
						if(y == 0 && z_surf_height > yres)
							*(buf_data + offset + adjusted_pitch*(z_surf_height-1)) = newColor;

						offset++;
					}

					pData++;
				}
				offset += adjusted_pitch - xres;
			}
		}
	}
	else
	if(z_options & IMG_TRUECOLOR)
	{
		if(z_options & IMG_ALPHA)
		{
			// if this image has a separate alpha channel,
			// combine the alpha and color data into a 32 bit surface texture for D3D to draw
			unsigned char *pDataOld;
			unsigned char *pAlphaOld;
			unsigned long *pDataNew;
			pDataOld = (unsigned char *) z_data; 
			pAlphaOld = (unsigned char *) z_alpha; 
			pDataNew = (unsigned long *) ddsd.lpSurface;
			const int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			register int offset = 0;

			for(int y = 0 ; y < z_yres ; y++)
			{
				offset = y * adjusted_pitch;
				for(int x = 0 ; x < z_xres ; x++)
				{
					if(x < xres && y < yres)
					{
						const int clr = ((*pDataOld)<<16) | ((*(pDataOld+1))<<8) | ((*(pDataOld+2))<<0);
						const int alpha = z_alpha ? *pAlphaOld : 0;
						const int newColor = clr + (alpha << 24);
						*(pDataNew + offset) = newColor;

						// clamp by one pixel to prevent oddities at scaled edges
						if(x == xres-1 && z_surf_width > xres)
							*(pDataNew + offset + 1) = newColor;
						if(y == yres-1 && z_surf_height > yres)
						{
							*(pDataNew + offset + adjusted_pitch) = newColor;
							if(x == xres-1 && z_surf_width > xres)
								*(pDataNew + offset + adjusted_pitch + 1) = newColor;
						}
						if(x == 0 && z_surf_width > xres)
							*(pDataNew + offset + z_surf_width-1) = newColor;
						if(y == 0 && z_surf_height > yres)
							*(pDataNew + offset + adjusted_pitch*(z_surf_height-1)) = newColor;

						offset++;
					}
					pDataOld += 3;
					pAlphaOld++;
				}
			}

		}
		else
		{
			if(z_surf_bpp == 24)
			{
				unsigned char * buf_data = (unsigned char *) ddsd.lpSurface;
				const int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
				const int adjusted_pitch = ddsd.lPitch / 3; // /3 for 24 bits
				unsigned char *pData = (unsigned char*) z_data;
				register int offset = 0;

				for(int y = 0 ; y < z_yres ; y++)
				{
					for(int x = 0 ; x < z_xres ; x++)
					{
						if(x < xres && y < yres)
						{
							const int red = *pData;
							const int green = *(pData+1);
							const int blue = *(pData+2);
							*(buf_data + offset) = red;
							*(buf_data + offset+1) = green;
							*(buf_data + offset+2) = blue;

							// clamp by one pixel to prevent oddities at scaled edges
							if(x == xres-1 && z_surf_width > xres)
							{
								*(buf_data + offset + 3) = red;
								*(buf_data + offset + 4) = green;
								*(buf_data + offset + 5) = blue;
							}
							if(y == yres-1 && z_surf_height > yres)
							{
								*(buf_data + offset + adjusted_pitch  ) = red;
								*(buf_data + offset + adjusted_pitch+1) = green;
								*(buf_data + offset + adjusted_pitch+2) = blue;
								if(x == xres-1 && z_surf_width > xres)
								{
									*(buf_data + offset + adjusted_pitch+3) = red;
									*(buf_data + offset + adjusted_pitch+4) = green;
									*(buf_data + offset + adjusted_pitch+5) = blue;
								}
							}
							if(x == 0 && z_surf_width > xres)
							{
								*(buf_data + offset + (z_surf_width-1)*3  ) = red;
								*(buf_data + offset + (z_surf_width-1)*3+1) = green;
								*(buf_data + offset + (z_surf_width-1)*3+2) = blue;
							}
							if(y == 0 && z_surf_height > yres)
							{
								*(buf_data + offset + adjusted_pitch*(z_surf_height-1)  ) = red;
								*(buf_data + offset + adjusted_pitch*(z_surf_height-1)+1) = green;
								*(buf_data + offset + adjusted_pitch*(z_surf_height-1)+2) = blue;
							}

							offset += 3;
						}

						pData += 3;
					}
					offset += adjusted_pitch - xres;
				}
			}
			else
			if(z_surf_bpp == 32)
			{
				unsigned long * buf_data = (unsigned long *) ddsd.lpSurface;
				const int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
				const int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
				unsigned char *pData = (unsigned char*) z_data;
				register int offset = 0;

				for(int y = 0 ; y < z_yres ; y++)
				{
					for(int x = 0 ; x < z_xres ; x++)
					{
						if(x < xres && y < yres)
						{
							const int clr = ((*pData)<<16) | ((*(pData+1))<<8) | ((*(pData+2))<<0);
							*(buf_data + offset) = clr;

							// clamp by one pixel to prevent oddities at scaled edges
							if(x == xres-1 && z_surf_width > xres)
							{
								*(buf_data + offset + 1) = clr;
							}
							if(y == yres-1 && z_surf_height > yres)
							{
								*(buf_data + offset + adjusted_pitch) = clr;
								if(x == xres-1 && z_surf_width > xres)
								{
									*(buf_data + offset + adjusted_pitch+1) = clr;
								}
							}
							if(x == 0 && z_surf_width > xres) *(buf_data + offset + z_surf_width-1) = clr;
							if(y == 0 && z_surf_height > yres) *(buf_data + offset + adjusted_pitch*(z_surf_height-1)) = clr;

							offset++;
						}

						pData += 3;
					}
					offset += adjusted_pitch - xres;
				}
			}
		}
	}
	else
	if(z_options & IMG_TRUECOLORX)
	{
		if(z_options & IMG_ALPHA)
		{
			// if this image has a separate alpha channel,
			// combine the alpha and color data into a 32 bit surface texture for D3D to draw
			unsigned long *pDataOld;
			unsigned char *pAlphaOld;
			unsigned long *pDataNew;
			pDataOld = (unsigned long *) z_data; 
			pAlphaOld = (unsigned char *) z_alpha; 
			pDataNew = (unsigned long *) ddsd.lpSurface;
			const int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			register int offset = 0;

			for(int y = 0 ; y < z_yres ; y++)
			{
				offset = y * adjusted_pitch;
				for(int x = 0 ; x < z_xres ; x++)
				{
					if(x < xres && y < yres)
					{
						const int clr = *pDataOld;
						const int alpha = z_alpha ? *pAlphaOld : 0;  // this should be 0 if the alpha is already in the color
						const int newColor = clr + (alpha << 24);
						*(pDataNew + offset) = newColor;

						// clamp by one pixel to prevent oddities at scaled edges
						if(x == xres-1 && z_surf_width > xres) *(pDataNew + offset + 1) = newColor;
						if(y == yres-1 && z_surf_height > yres)
						{
							*(pDataNew + offset + adjusted_pitch) = newColor;
							if(x == xres-1 && z_surf_width > xres) *(pDataNew + offset + adjusted_pitch + 1) = newColor;
						}
						if(x == 0 && z_surf_width > xres) *(pDataNew + offset + z_surf_width-1) = newColor;
						if(y == 0 && z_surf_height > yres) *(pDataNew + offset + adjusted_pitch*(z_surf_height-1)) = newColor;

						offset++;
					}

					pDataOld++;
					pAlphaOld++;
				}
			}
		}
		else
		{
			unsigned long * buf_data = (unsigned long *) ddsd.lpSurface;
			const int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
			const int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			unsigned long *pData = (unsigned long*) z_data;
			register int offset = 0;

			for(int y = 0 ; y < z_yres ; y++)
			{
				for(int x = 0 ; x < z_xres ; x++)
				{
					if(x < xres && y < yres)
					{
						const int newColor = *pData;
						*(buf_data + offset) = newColor;

						// clamp by one pixel to prevent oddities at scaled edges
						if(x == xres-1 && z_surf_width > xres) *(buf_data + offset + 1) = newColor;
						if(y == yres-1 && z_surf_height > yres)
						{
							*(buf_data + offset + adjusted_pitch) = newColor;
							if(x == xres-1 && z_surf_width > xres) *(buf_data + offset + adjusted_pitch + 1) = newColor;
						}
						if(x == 0 && z_surf_width > xres) *(buf_data + offset + z_surf_width-1) = newColor;
						if(y == 0 && z_surf_height > yres) *(buf_data + offset + adjusted_pitch*(z_surf_height-1)) = newColor;

						offset++;
					}

					pData++;
				}
				offset += adjusted_pitch - xres;
			}
		}
	}

 	z_status = z_surface->Unlock(NULL); // unlock drawing surface
	if(z_status != DD_OK){ 	
		GameX.z_DirectDrawError(z_status);
		return;
	}

	if(GameX.z_CanDeleteMData() && z_data != NULL)
	{
		if(!(z_usage & LOAD_NON3D) || !((z_usage & LOAD_ALPHA && !(z_usage & LOAD_MASKED))))
		{
			DeleteData();
		}
	}
}

inline int ImageX::GetBytesPerPixel (ImageOps ops) const
{												// Pixel Format
	if(ops & IMG_MERGEALPHA) {					// I R G B A
		if(ops & IMG_INDEXED) return 2;		// 8-0-0-0-8
		if(ops & IMG_LOWCOLOR) return 2;		// 0-4-4-4-4
		if(ops & IMG_HIGHCOLOR) return 3;		// 0-6-5-5-8
		if(ops & IMG_TRUECOLOR) return 4;		// 0-8-8-8-8
		if(ops & IMG_TRUECOLORX) return 4;		// 0-8-8-8-8
	} else
	{
		if(ops & IMG_INDEXED) return 1;		// 8-0-0-0-0
		if(ops & IMG_LOWCOLOR) return IMG_UNDEF;// undefined
		if(ops & IMG_HIGHCOLOR) return 2;		// 0-6-5-5-0
		if(ops & IMG_TRUECOLOR) return 3;		// 0-8-8-8-0
		if(ops & IMG_TRUECOLORX) return 4;		// 0-8-8-8-8
	}
	return IMG_UNDEF;
}

inline int ImageX::GetFormat (ImageOps ops) const
{
	if(ops & IMG_INDEXED) return IMG_INDEXED;
	if(ops & IMG_LOWCOLOR) return IMG_LOWCOLOR;
	if(ops & IMG_HIGHCOLOR) return IMG_HIGHCOLOR;
	if(ops & IMG_TRUECOLOR) return IMG_TRUECOLOR;
	if(ops & IMG_TRUECOLORX) return IMG_TRUECOLORX;
	return IMG_UNDEF;
}

inline int ImageX::GetAlpha (ImageOps ops) const
{	
	if(ops & IMG_ALPHA) return IMG_ALPHA;
	return IMG_UNDEF;
}

inline void ImageX::TurnOff (ImageOps opt)
{	
	if(z_options & opt)
		z_options ^= opt;
}

inline void ImageX::TurnOn (ImageOps opt)
{	
	z_options |= opt;
}

void ImageX::Size (int xr, int yr, ImageOps ops)
{
	if(GetFormat(ops) == IMG_UNDEF)
	{
		if(debug.IsRunningDebug())
			GameX.z_SafeMessageBox("Error: Image options passed to ImageX::Size() must include a format such as IMG_HIGHCOLOR", "GameX Error",MB_OK|MB_ICONSTOP) ;
		return;
	}

	int bpp = GetBytesPerPixel (ops);
	
	if(bpp != -1)
	{
		SAFE_DELETE_ARRAY(z_data);	
		z_data = new unsigned char[xr * yr * bpp];
	}
	if(ops & IMG_ALPHA)
	{
		SAFE_DELETE_ARRAY(z_alpha);
		z_alpha = new unsigned char[xr * yr];
	}	
	z_xres = xr; z_yres = yr; z_size = xr * yr;
	z_options = ops;
}

void ImageX::Fill (int r, int g, int b, int a) // r, g, b, a are out of 255
{
	if(z_data != NULL)
	{
		unsigned short * pData = (unsigned short*) z_data;
		int size = z_size * GetBytesPerPixel(z_options) / sizeof(unsigned short);
		for(int n = 0; n < size; n++)
		{
			*pData++ = (((r >> 3) << GameX.z_win_shiftr) + ((g >> 2) << GameX.z_win_shiftg) + ((b >> 3) /*<< GameX.z_win_shiftb*/));
		}
	}

	if(z_alpha != NULL)
	{
		unsigned char * pAlpha = (unsigned char*) z_alpha;
		for(int n = 0; n < z_size; n++)
		{
			*pAlpha++ = a;
		}
	}

	if(z_surface != NULL)
	{
		GameX.AccessPixels(this);
		for(int y = z_fy1 ; y < z_fy2 ; y++)
			for(int x = z_fx1 ; x < z_fx2 ; x++)
				GameX.DrawPixel(x,y,r,g,b,a);
		GameX.EndPixelAccess();
	}
	ResetFilterRect();
}

void ImageX::SwapColors (ColorSwapType type) // rotates or swaps the image's color values
{
	if(type < 0) return;
	type %= 20;
	switch(type) { // if it's a complicated LAB color change:
		case COLOR_SWAP_BA:   ChangeColors(COLOR_SET_B,COLOR_SET_A); return;
		case COLOR_SWAP_NBA:  ChangeColors(COLOR_SET_B,COLOR_SET_A,-1.0f,1.0f); return;
		case COLOR_SWAP_BNA:  ChangeColors(COLOR_SET_B,COLOR_SET_A,1.0f,-1.0f); return;
		case COLOR_SWAP_NBNA: ChangeColors(COLOR_SET_B,COLOR_SET_A,-1.0f,-1.0f); return;
		case COLOR_SWAP_ANB:  ChangeColors(COLOR_SET_A,COLOR_SET_B,1.0f,-1.0f); return;
		case COLOR_SWAP_NAB:  ChangeColors(COLOR_SET_A,COLOR_SET_B,-1.0f,1.0f); return;
		case COLOR_SWAP_NANB: ChangeColors(COLOR_SET_A,COLOR_SET_B,-1.0f,-1.0f); return;
		case COLOR_SWAP_BB:   ChangeColors(COLOR_SET_B,COLOR_SET_B); return;
		case COLOR_SWAP_NBB:  ChangeColors(COLOR_SET_B,COLOR_SET_B, -1.0f,1.0f); return;
		case COLOR_SWAP_BNB:  ChangeColors(COLOR_SET_B,COLOR_SET_B, 1.0f,-1.0f); return;
		case COLOR_SWAP_NBNB: ChangeColors(COLOR_SET_B,COLOR_SET_B, -1.0f,-1.0f); return;
		case COLOR_SWAP_AA:   ChangeColors(COLOR_SET_A,COLOR_SET_A); return;
		case COLOR_SWAP_NAA:  ChangeColors(COLOR_SET_A,COLOR_SET_A, -1.0f,1.0f); return;
		case COLOR_SWAP_ANA:  ChangeColors(COLOR_SET_A,COLOR_SET_A, 1.0f,-1.0f); return;
		case COLOR_SWAP_NANA: ChangeColors(COLOR_SET_A,COLOR_SET_A, -1.0f,-1.0f); return;
	}

	// otherwise, plain color component swapping:
	
	GameX.AccessPixels(this);
	for(int y = z_fy1 ; y < z_fy2 ; y++)
	{
		if(z_ymult != 1.0f && y != z_fy1 && (int)(z_ymult * (float)(y-1)) == (int)(z_ymult * (float)(y))) continue;
		for(int x = z_fx1 ; x < z_fx2 ; x++)
		{
			if(z_xmult != 1.0f && x != z_fx1 && (int)(z_xmult * (float)(x-1)) == (int)(z_xmult * (float)(x))) continue;
			int r,g,b,a;
			GameX.ReadPixel(x,y,r,g,b,a);
			if(r == 0 && g == 0 && b == 0) continue;
			switch(type)
			{
				case COLOR_SWAP_ROTATE120: GameX.DrawPixel(x,y,b,r,g,a); break; // rotate 120 degrees
				case COLOR_SWAP_ROTATE240: GameX.DrawPixel(x,y,g,b,r,a); break; // rotate 240 degrees
				case COLOR_SWAP_REDGREEN:  GameX.DrawPixel(x,y,g,r,b,a); break; // swap red and green
				case COLOR_SWAP_REDBLUE:   GameX.DrawPixel(x,y,b,g,r,a); break; // swap red and blue
				case COLOR_SWAP_GREENBLUE: GameX.DrawPixel(x,y,r,b,g,a); break; // swap green and blue
			}
		}
	}
	GameX.EndPixelAccess();
	ResetFilterRect();
}


void ImageX::ChangeColors(ColorSetType a_settype, ColorSetType b_settype, float a_set_mult, float b_set_mult, float a_app_alpha, float b_app_alpha, ColorSetCondition a_acond, ColorSetCondition a_bcond, ColorSetCondition b_acond, ColorSetCondition b_bcond)
{
 	int types [2], conditions [4];
	float mult [2], alpha [2];
	types[0] = a_settype;
	types[1] = b_settype;
	mult[0] = a_set_mult;
	mult[1] = b_set_mult;
	alpha[0] = a_app_alpha;
	alpha[1] = b_app_alpha;
	conditions[0] = a_acond;
	conditions[1] = a_bcond;
	conditions[2] = b_acond;
	conditions[3] = b_bcond;
	GameX.AccessPixels(this);
	for(int y = z_fy1 ; y < z_fy2 ; y++)
	{
		if(z_ymult != 1.0f && y != z_fy1 && (int)(z_ymult * (float)(y-1)) == (int)(z_ymult * (float)(y))) continue;
		for(int x = z_fx1 ; x < z_fx2 ; x++)
		{
			if(z_xmult != 1.0f && x != z_fx1 && (int)(z_xmult * (float)(x-1)) == (int)(z_xmult * (float)(x))) continue;
			int r,g,b,a;
			GameX.ReadPixel(x,y,r,g,b,a);
			if(r == 0 && g == 0 && b == 0) continue;
			int L, c[2], o[2], match [2]; // c for color, o for original
			ConvertRGBToLab(r,g,b,L,o[0],o[1]);
			c[0] = o[0];
			c[1] = o[1];
			match[0] = 0;
			match[1] = 0;
			for(int s = 0 ; s < 2 ; s++) { // setting match
				for(int t = 0 ; t < 2 ; t++) { // being tested
					switch(conditions[s + t+t])
					{
						case COLOR_COND_NEGATIVE: if(o[s] < 0) match[t]++; break; // < 0
						case COLOR_COND_POSITIVE: if(o[s] > 0) match[t]++; break; // > 0
						case COLOR_COND_ANYTHING: match[t]++; break; // anything
						case COLOR_COND_NEUTRAL:  if(o[s] > -50 && o[s] < 50) match[t]++; break; // -50 to 50
						case COLOR_COND_GREATER: if(o[s] > o[1-s]) match[t]++; break; // > other
						case COLOR_COND_LESS: if(o[s] < o[1-s]) match[t]++; break; // < other
						case COLOR_COND_NEAR: if(o[s] > o[1-s]-50 && o[s] < o[1-s]+50) match[t]++; break; // other-50 to other+50
					}
				}
			}
			if(match[0] < 2 && match[1] < 2) continue;
			for(int t = 0 ; t < 2 ; t++)
			{
				if(match[t] < 2) continue;
				switch(types[t])
				{
					case COLOR_SET_CONSTANT: c[t] = 50; break;
					case COLOR_SET_A: c[t] = o[0]; break;
					case COLOR_SET_B: c[t] = o[1]; break;
					case COLOR_SET_ABS_A: c[t] = abs(o[0]); break;
					case COLOR_SET_ABS_B: c[t] = abs(o[1]); break;
					case COLOR_SET_MAX: c[t] = max(o[0],o[1]); break;
					case COLOR_SET_MIN: c[t] = min(o[0],o[1]); break;
					case COLOR_SET_A_PLUS_B: c[t] = o[0]+o[1]; break;
					case COLOR_SET_A_MINUS_B: c[t] = o[0]-o[1]; break;
					case COLOR_SET_ABS_A_PLUS_ABS_B: c[t] = abs(o[0])+abs(o[1]); break;
					case COLOR_SET_ABS_A_MINUS_ABS_B: c[t] = abs(o[0])-abs(o[1]); break;
					case COLOR_SET_L: c[t] = Round((float)(L*L*L)/(100.0f*100.0f)); break;
					case COLOR_SET_L_SMOOTH: c[t] = Round(((float)(L*L*L)/(100.0f*100.0f)-50)*2); break;
					case COLOR_SET_SIN_A: c[t] = Round(sinf((float)o[0]*((360.0f/100.0f)*DEGtoRAD))*100.0f); break;
					case COLOR_SET_SIN_B: c[t] = Round(sinf((float)o[1]*((360.0f/100.0f)*DEGtoRAD))*100.0f); break;
					case COLOR_SET_COS_A: c[t] = Round(cosf((float)o[0]*((360.0f/100.0f)*DEGtoRAD))*100.0f); break;
					case COLOR_SET_COS_B: c[t] = Round(cosf((float)o[1]*((360.0f/100.0f)*DEGtoRAD))*100.0f); break;
					case COLOR_SET_SIN_L: c[t] = Round(sinf((float)(L*L*L)/(100.0f*100.0f)*((360.0f/100.0f)*DEGtoRAD))*100.0f); break;
					case COLOR_SET_COS_L: c[t] = Round(cosf((float)(L*L*L)/(100.0f*100.0f)*((360.0f/100.0f)*DEGtoRAD))*100.0f); break;
				}
				if(mult[t] != 1.0f)
				{
					if(mult[t] == 0.0f)
						c[t] = 0;
					else if(mult[t] == -1.0f)
						c[t] = -c[t];
					else
						c[t] = Round((float)c[t] * mult[t]);
				}
				if(alpha[t] != 1.0f)
				{
					if(alpha[t] == 0.0f)
						c[t] = o[t];
					else
						c[t] = Round((float)c[t]*alpha[t] + (float)o[t]*(1.0f-alpha[t]));
				}
			}
			ConvertLabToRGB(L,c[0],c[1],r,g,b);
			GameX.DrawPixel(x,y,r,g,b,a);
		}
	}
	GameX.EndPixelAccess();
	ResetFilterRect();
}

void ImageX::Invert (bool preserve_color) // inverts the image
{
	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;
	GameX.AccessPixels(this);
	int maxl = 1000;
	for(int y = z_fy1 ; y < z_fy2 ; y++)
		for(int x = z_fx1 ; x < z_fx2 ; x++)
		{
			int r,g,b,a;
			GameX.ReadPixel(x,y,r,g,b,a);

			if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
				continue;

			if(z_usage & LOAD_MASKED && r>255-rmin && g>255-gmin && b>255-bmin)
			{
				r = 255-rmin;
				g = 255-gmin;
				b = 255-bmin;
			}
			if(preserve_color)
			{
				int L,A,B;
				ConvertRGBToLab(r,g,b,L,A,B);
				ConvertLabToRGB(150-L,A,B,r,g,b);
				GameX.DrawPixel(x,y,r,g,b,a);
			} else
				GameX.DrawPixel(x,y,255-r,255-g,255-b,a);
		}
	GameX.EndPixelAccess();
	ResetFilterRect();
}

void ImageX::ChangeBrightness (int brightness, float contrast)
{
	ChangeBrightnesses(brightness,brightness,brightness,0,contrast,contrast,contrast,1.0f);
}

void ImageX::ChangeBrightnesses (int brightness_r, int brightness_g, int brightness_b, int brightness_a, float contrast_r, float contrast_g, float contrast_b, float contrast_a) // changes brightness and contrast
{
	int offset_r = brightness_r + (contrast_r ? Round(255.0f*(1.0f - contrast_r)/2.0f) : 0);
	int offset_g = brightness_g + (contrast_g ? Round(255.0f*(1.0f - contrast_g)/2.0f) : 0);
	int offset_b = brightness_b + (contrast_b ? Round(255.0f*(1.0f - contrast_b)/2.0f) : 0);
	int offset_a = brightness_a + (contrast_a ? Round(255.0f*(1.0f - contrast_a)/2.0f) : 0);
	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;
	GameX.AccessPixels(this);
	for(int y = z_fy1 ; y < z_fy2 ; y++)
	{
		if(z_ymult != 1.0f && y != z_fy1 && (int)(z_ymult * (float)(y-1)) == (int)(z_ymult * (float)(y))) continue;
		for(int x = z_fx1 ; x < z_fx2 ; x++)
		{
			if(z_xmult != 1.0f && x != z_fx1 && (int)(z_xmult * (float)(x-1)) == (int)(z_xmult * (float)(x))) continue;
			int r,g,b,a;
			GameX.ReadPixel(x,y,r,g,b,a);

			if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
				continue;

			if(contrast_r != 1.0f) r = Round((float)r * contrast_r);
			if(contrast_g != 1.0f) g = Round((float)g * contrast_g);
			if(contrast_b != 1.0f) b = Round((float)b * contrast_b);
			if(contrast_a != 1.0f) a = Round((float)a * contrast_a);
			r += offset_r;
			g += offset_g;
			b += offset_b;
			a += offset_a;

			if(r > 255) r = 255;
			if(g > 255) g = 255;
			if(b > 255) b = 255;
			if(a > 255) a = 255;
			if(r < 0) r = 0;
			if(g < 0) g = 0;
			if(b < 0) b = 0;
			if(a < 0) a = 0;

			if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
				b = bmin;

			GameX.DrawPixel(x,y,r,g,b,a);
		}
	}
	GameX.EndPixelAccess();
	ResetFilterRect();
}

void ImageX::ChangeSaturation (float sat)
{
	ChangeSaturations(sat,sat,sat);
}

void ImageX::ChangeSaturations (float sat_r, float sat_g, float sat_b) // reduces or enhances color while keeping brightness the same
{
	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;
	GameX.AccessPixels(this);
	for(int y = z_fy1 ; y < z_fy2 ; y++)
	{
		if(z_ymult != 1.0f && y != z_fy1 && (int)(z_ymult * (float)(y-1)) == (int)(z_ymult * (float)(y))) continue;
		for(int x = z_fx1 ; x < z_fx2 ; x++)
		{
			if(z_xmult != 1.0f && x != z_fx1 && (int)(z_xmult * (float)(x-1)) == (int)(z_xmult * (float)(x))) continue;
			int r,g,b,a;
			GameX.ReadPixel(x,y,r,g,b,a);

			if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
				continue;
			
			float red, green, blue;
			red = (float)r/255.0f;
			green = (float)g/255.0f;
			blue = (float)b/255.0f;
			float gray = red*RED_LUMINANCE + green*GREEN_LUMINANCE + blue*BLUE_LUMINANCE;
			
			red = (red*sat_r + gray*(1.0f-sat_r));
			green = (green*sat_g + gray*(1.0f-sat_g));
			blue = (blue*sat_b + gray*(1.0f-sat_b));
			
			r = Round(red*255.0f);
			g = Round(green*255.0f);
			b = Round(blue*255.0f);

			if(r > 255) r = 255;
			if(g > 255) g = 255;
			if(b > 255) b = 255;
			if(r < 0) r = 0;
			if(g < 0) g = 0;
			if(b < 0) b = 0;

			if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
			{
				r = rmin;
				g = gmin;
				b = bmin;
			}

			GameX.DrawPixel(x,y,r,g,b,a);
		}
	}
	GameX.EndPixelAccess();
	ResetFilterRect();
}

void ImageX::ConvertToGrayscale ()
{
	ChangeSaturation(0.0f);
}

void ImageX::SetChannel (int channel, int value) // sets all of one channel to one value
{
	channel %= 4;
	if(value < 0) value = 0;
	if(value > 255) value = 255;
	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;
	GameX.AccessPixels(this);
	for(int y = z_fy1 ; y < z_fy2 ; y++)
	for(int x = z_fx1 ; x < z_fx2 ; x++)
	{
		int r,g,b,a;
		GameX.ReadPixel(x,y,r,g,b,a);
		
		if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
			continue;
		
		switch(channel)
		{
			case 0: r = value; break;
			case 1: g = value; break;
			case 2: b = value; break;
			case 3: a = value; break;
		}
		
		if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
		{
			if(channel == 0) r = rmin;
			if(channel == 1) g = gmin;
			if(channel == 2) b = bmin;
		}
		
		GameX.DrawPixel(x,y,r,g,b,a);
	}
	GameX.EndPixelAccess();
	ResetFilterRect();
}

void ImageX::AddNoise (int noise)
{
	AddNoise(noise,noise,noise,0);
}

void ImageX::AddNoise (int noise_r, int noise_g, int noise_b, int noise_a) // adds random noise to the image
{
	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;
z_pushDrawing();
	GameX.AccessPixels(this);
	for(int y = z_fy1 ; y < z_fy2 ; y++)
	{
		if(z_ymult != 1.0f && y != z_fy1 && (int)(z_ymult * (float)(y-1)) == (int)(z_ymult * (float)(y))) continue;
		for(int x = z_fx1 ; x < z_fx2 ; x++)
		{
			if(z_xmult != 1.0f && x != z_fx1 && (int)(z_xmult * (float)(x-1)) == (int)(z_xmult * (float)(x))) continue;
			int r,g,b,a;
			GameX.ReadPixel(x,y,r,g,b,a);

			if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
				continue;

			if(noise_r < 0) r += GameX.GetRandomInt(noise_r,0);
			else if(noise_r > 0) r += GameX.GetRandomInt(-noise_r,noise_r);
			if(noise_g < 0) g += GameX.GetRandomInt(noise_g,0);
			else if(noise_r > 0) g += GameX.GetRandomInt(-noise_g,noise_g);
			if(noise_b < 0) b += GameX.GetRandomInt(noise_b,0);
			else if(noise_r > 0) b += GameX.GetRandomInt(-noise_b,noise_b);
			if(noise_a < 0) a += GameX.GetRandomInt(noise_a,0);
			else if(noise_a > 0) a += GameX.GetRandomInt(-noise_a,noise_a);

			if(r > 255) r = 255;
			if(g > 255) g = 255;
			if(b > 255) b = 255;
			if(a > 255) a = 255;
			if(r < 0) r = 0;
			if(g < 0) g = 0;
			if(b < 0) b = 0;
			if(a < 0) a = 0;

			if(z_usage & LOAD_MASKED && r<rmin && g<gmin && b<bmin)
				b = bmin;

			GameX.DrawPixel(x,y,r,g,b,a);
		}
	}
	GameX.EndPixelAccess();
z_popDrawing();
	ResetFilterRect();
}

void ImageX::Blur (float radius)
{
	Blur(radius,radius,radius,radius);
}

void ImageX::HorzBlur (float radius)
{
	HorzBlur(radius,radius,radius,radius);
}

void ImageX::VertBlur (float radius)
{
	VertBlur(radius,radius,radius,radius);
}


void ImageX::HorzBlur (float red_radius, float green_radius, float blue_radius, float alpha_radius)
{
	int xWidth = z_fx2-z_fx1;
	int size = (xWidth)*(z_fy2-z_fy1)*4;
	unsigned char * data = new unsigned char[size];

	GameX.AccessPixels(this);
	for(int y1 = z_fy1 ; y1 < z_fy2 ; y1++)
	{
		if(z_ymult != 1.0f && y1 != z_fy1 && (int)(z_ymult * (float)(y1-1)) == (int)(z_ymult * (float)(y1))) continue;
		for(int x1 = z_fx1 ; x1 < z_fx2 ; x1++)
		{
			if(z_xmult != 1.0f && x1 != z_fx1 && (int)(z_xmult * (float)(x1-1)) == (int)(z_xmult * (float)(x1))) continue;
			int c[4];
			GameX.ReadPixel(x1,y1,c[0],c[1],c[2],c[3]);
			for(int i = 0 ; i < 4 ; i++)
				data[i+4*(x1-z_fx1+(xWidth)*(y1-z_fy1))] = (unsigned char)c[i];
		}
	}
	
	float rMaxDistSq = red_radius*red_radius*4.0f;
	float gMaxDistSq = green_radius*green_radius*4.0f;
	float bMaxDistSq = blue_radius*blue_radius*4.0f;
	float aMaxDistSq = alpha_radius*alpha_radius*4.0f;
	float maxDistSq = max(rMaxDistSq,max(gMaxDistSq,max(bMaxDistSq,aMaxDistSq)));
	int radInt = Round(ceilf(sqrtf(maxDistSq)));

	float est = (float)(radInt)*(float)size;

	if(est > 32000000.0f) { // cap the blurring time at estimated 32x10^6 loops
		est = 32000000.0f/est;
		rMaxDistSq *= est;
		gMaxDistSq *= est;
		bMaxDistSq *= est;
		aMaxDistSq *= est;
		maxDistSq = max(rMaxDistSq,max(gMaxDistSq,max(bMaxDistSq,aMaxDistSq)));
		radInt = Round(ceilf(sqrtf(maxDistSq)));
	}

	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;

	float diffusion = 0.0f;
	for(int y = z_fy1 ; y < z_fy2 ; y++)
	{
		if(z_ymult != 1.0f && y != z_fy1 && (int)(z_ymult * (float)(y-1)) == (int)(z_ymult * (float)(y))) continue;
		for(int x = z_fx1 ; x < z_fx2 ; x++)
{

			diffusion += 0.33f;
			if(diffusion > 4.0f)
				diffusion -= 4.0f;

			if(z_xmult != 1.0f && x != z_fx1 && (int)(z_xmult * (float)(x-1)) == (int)(z_xmult * (float)(x))) continue;
			int offset = 4*((x-z_fx1)+(xWidth)*(y-z_fy1));
			float rSum = 0, gSum = 0, bSum = 0, aSum = 0; // sum of color values read times current amount
			float rTot = 0, gTot = 0, bTot = 0, aTot = 0; // sum of current amounts
			for(int x2 = x-radInt ; x2 <= x+radInt ; x2++)
			{
				float xdiff = (float)(x2-x);
				float distSq = xdiff*xdiff;
				if(distSq > maxDistSq+0.001f)
					continue;
				if(distSq < 0.005f)
					distSq = 1.0f;
				if(xdiff < 0) xdiff = -xdiff;
				if(xdiff < 0.005f)
					xdiff = 1.0f;

				int x3 = x2;
				if(x3< z_fx1) x3 += z_fx2-z_fx1;
				if(x3 >= z_fx2) x3 -= z_fx2-z_fx1;
				int r,g,b,a;
				GameX.ReadPixel(x3,y,r,g,b,a);

				if(a < gmin || (z_usage&LOAD_MASKED && (r<rmin && g<gmin && b<bmin)))
				{
					r = (int)data[0+offset];
					g = (int)data[1+offset];
					b = (int)data[2+offset];
				}
				if(distSq < rMaxDistSq+0.001f)
				{
					float r_amount = (float)red_radius / xdiff;
					rTot += r_amount;
					if(r >= rmin) rSum += (float)r * r_amount;
				}
				if(distSq < gMaxDistSq+0.001f)
				{
					float g_amount = (float)green_radius / xdiff;
					gTot += g_amount;
					if(g >= gmin) gSum += (float)g * g_amount;
				}
				if(distSq < bMaxDistSq+0.001f)
				{
					float b_amount = (float)blue_radius / xdiff;
					bTot += b_amount;
					if(b >= bmin) bSum += (float)b * b_amount;
				}
				if(distSq < aMaxDistSq+0.001f)
				{
					float a_amount = (float)alpha_radius / xdiff;
					aTot += a_amount;
					if(a >= gmin) aSum += (float)a * a_amount;
				}
			}
			if(rTot) data[0+offset] = (unsigned char)(rSum/rTot);
			if(gTot) data[1+offset] = (unsigned char)(gSum/gTot);
			if(bTot) data[2+offset] = (unsigned char)(bSum/bTot);
			if(aTot) data[3+offset] = (unsigned char)(aSum/aTot);
		}
	}
	for(int y3 = z_fy1 ; y3 < z_fy2 ; y3++)
	{
		for(int x3 = z_fx1 ; x3 < z_fx2 ; x3++)
		{
			int offset = 4*(x3-z_fx1+xWidth*(y3-z_fy1));
			GameX.DrawPixel(x3,y3,(int)data[0+offset],(int)data[1+offset],(int)data[2+offset],(int)data[3+offset]);
		}
	}
	GameX.EndPixelAccess();

	SAFE_DELETE_ARRAY(data);

	ResetFilterRect();
}

void ImageX::VertBlur (float red_radius, float green_radius, float blue_radius, float alpha_radius)
{
	int xWidth = z_fx2-z_fx1;
	int size = (xWidth)*(z_fy2-z_fy1)*4;
	unsigned char * data = new unsigned char[size];

	GameX.AccessPixels(this);
	for(int y1 = z_fy1 ; y1 < z_fy2 ; y1++)
	{
		if(z_ymult != 1.0f && y1 != z_fy1 && (int)(z_ymult * (float)(y1-1)) == (int)(z_ymult * (float)(y1))) continue;
		for(int x1 = z_fx1 ; x1 < z_fx2 ; x1++)
		{
			if(z_xmult != 1.0f && x1 != z_fx1 && (int)(z_xmult * (float)(x1-1)) == (int)(z_xmult * (float)(x1))) continue;
			int c[4];
			GameX.ReadPixel(x1,y1,c[0],c[1],c[2],c[3]);
			for(int i = 0 ; i < 4 ; i++)
				data[i+4*(x1-z_fx1+(xWidth)*(y1-z_fy1))] = (unsigned char)c[i];
		}
	}
	
	float rMaxDistSq = red_radius*red_radius*4.0f;
	float gMaxDistSq = green_radius*green_radius*4.0f;
	float bMaxDistSq = blue_radius*blue_radius*4.0f;
	float aMaxDistSq = alpha_radius*alpha_radius*4.0f;
	float maxDistSq = max(rMaxDistSq,max(gMaxDistSq,max(bMaxDistSq,aMaxDistSq)));
	int radInt = Round(ceilf(sqrtf(maxDistSq)));

	float est = (float)(radInt)*(float)size;

	if(est > 32000000.0f) { // cap the blurring time at estimated 32x10^6 loops
		est = 32000000.0f/est;
		rMaxDistSq *= est;
		gMaxDistSq *= est;
		bMaxDistSq *= est;
		aMaxDistSq *= est;
		maxDistSq = max(rMaxDistSq,max(gMaxDistSq,max(bMaxDistSq,aMaxDistSq)));
		radInt = Round(ceilf(sqrtf(maxDistSq)));
	}

	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;

	for(int y = z_fy1 ; y < z_fy2 ; y++)
	{
		if(z_ymult != 1.0f && y != z_fy1 && (int)(z_ymult * (float)(y-1)) == (int)(z_ymult * (float)(y))) continue;
		for(int x = z_fx1 ; x < z_fx2 ; x++)
		{
			if(z_xmult != 1.0f && x != z_fx1 && (int)(z_xmult * (float)(x-1)) == (int)(z_xmult * (float)(x))) continue;
			int offset = 4*((x-z_fx1)+(xWidth)*(y-z_fy1));
			float rSum = 0, gSum = 0, bSum = 0, aSum = 0; // sum of color values read times current amount
			float rTot = 0, gTot = 0, bTot = 0, aTot = 0; // sum of current amounts
			for(int y2 = y-radInt ; y2 <= y+radInt ; y2++)
{
///				if(y2<z_fy1||y2 >= z_fy2)
///					continue;
				float ydiff = (float)(y2-y);
				float distSq = ydiff*ydiff;
				if(distSq > maxDistSq+0.001f)
					continue;
				if(distSq < 0.005f)
					distSq = 1.0f;
				if(ydiff < 0) ydiff = -ydiff;
				if(ydiff < 0.005f)
					ydiff = 1.0f;

				int y3 = y2;
				if(y3< z_fy1) y3 += z_fy2-z_fy1;
				if(y3 >= z_fy2) y3 -= z_fy2-z_fy1;
				int r,g,b,a;
				GameX.ReadPixel(x,y3,r,g,b,a);

				if(a < gmin || (z_usage&LOAD_MASKED && (r<rmin && g<gmin && b<bmin)))
				{
					r = (int)data[0+offset];
					g = (int)data[1+offset];
					b = (int)data[2+offset];
				}
				if(distSq < rMaxDistSq+0.001f)
				{
					float r_amount = (float)red_radius / ydiff;
					rTot += r_amount;
					if(r >= rmin) rSum += (float)r * r_amount;
				}
				if(distSq < gMaxDistSq+0.001f)
				{
					float g_amount = (float)green_radius / ydiff;
					gTot += g_amount;
					if(g >= gmin) gSum += (float)g * g_amount;
				}
				if(distSq < bMaxDistSq+0.001f)
				{
					float b_amount = (float)blue_radius / ydiff;
					bTot += b_amount;
					if(b >= bmin) bSum += (float)b * b_amount;
				}
				if(distSq < aMaxDistSq+0.001f)
				{
					float a_amount = (float)alpha_radius / ydiff;
					aTot += a_amount;
					if(a >= gmin) aSum += (float)a * a_amount;
				}
			}
			if(rTot) data[0+offset] = (unsigned char)(rSum/rTot);
			if(gTot) data[1+offset] = (unsigned char)(gSum/gTot);
			if(bTot) data[2+offset] = (unsigned char)(bSum/bTot);
			if(aTot) data[3+offset] = (unsigned char)(aSum/aTot);
		}
	}
	for(int y3 = z_fy1 ; y3 < z_fy2 ; y3++)
	{
		for(int x3 = z_fx1 ; x3 < z_fx2 ; x3++)
		{
			int offset = 4*(x3-z_fx1+xWidth*(y3-z_fy1));
			GameX.DrawPixel(x3,y3,(int)data[0+offset],(int)data[1+offset],(int)data[2+offset],(int)data[3+offset]);
		}
	}
	GameX.EndPixelAccess();

	SAFE_DELETE_ARRAY(data);

	ResetFilterRect();
}

void ImageX::Blur (float red_radius, float green_radius, float blue_radius, float alpha_radius)
{
	// filter separably
	HorzBlur(red_radius, green_radius, blue_radius, alpha_radius);
	VertBlur(red_radius, green_radius, blue_radius, alpha_radius);
}

void ImageX::CopyTo (ImageX * dest) const
{
	if(dest == NULL || dest == VIEWPORT || z_status != IMG_OK) return;
	if(dest->z_status != IMG_OK) // if copying to an image that hasn't been loaded or created yet,
		dest->z_CreateWithUse(z_xres,z_yres,z_usage); // create it before copying into it

	int sy1 = z_fy1;
	int sy2 = z_fy2;
	int sx1 = z_fx1;
	int sx2 = z_fx2;
	int dy1 = dest->z_fy1;
	int dy2 = dest->z_fy2;
	int dx1 = dest->z_fx1;
	int dx2 = dest->z_fx2;
///	int x2 = (z_fx2 >= dest->z_xres) ? dest->z_surf_width : z_fx2;

	int sxWidth = sx2-sx1;
	int syHeight = sy2-sy1;

	// allocate enough data to hold source pixels matching (in size) the destination:
	int dxWidth = dx2-dx1;
	int dyHeight = dy2-dy1;
	int size = (dxWidth)*(dyHeight)*4;
	unsigned char * data = new unsigned char[size];

	// make a temporary copy of this image, scaled to destination size if necessary:
	GameX.AccessPixels(this);
	int c[4];
	if(sxWidth == dxWidth && syHeight == dyHeight) { // if no scaling needed:
		for(int y = dy1 ; y < dy2 ; y++)
		for(int x = dx1 ; x < dx2 ; x++)
		{
			GameX.ReadPixel(x,y,c[0],c[1],c[2],c[3]);
			for(int i = 0 ; i < 4 ; i++)
				data[i+4*(x-dx1+(dxWidth)*(y-dy1))] = (unsigned char)c[i];
		}
	} else { // otherwise scale with nearest-neighbor method:
		double xmult = (double)sxWidth/(double)dxWidth;
		double ymult = (double)syHeight/(double)dyHeight;
		for(int y = dy1 ; y < dy2 ; y++)
		{
			int gy = (int)((double)y*ymult);
			if(z_ymult != 1.0f && gy != z_fy1 && (int)(z_ymult * (float)(gy-1)) == (int)(z_ymult * (float)(gy))) continue;
			for(int x = dx1 ; x < dx2 ; x++)
			{
				int gx = (int)((double)x*xmult);
				if(z_xmult != 1.0f && gx != z_fx1 && (int)(z_xmult * (float)(gx-1)) == (int)(z_xmult * (float)(gx))) continue;
				GameX.ReadPixel(gx,gy,c[0],c[1],c[2],c[3]);
				for(int i = 0 ; i < 4 ; i++)
					data[i+4*(x-dx1+(dxWidth)*(y-dy1))] = (unsigned char)c[i];
			}
		}
	}
	GameX.EndPixelAccess();

	// now transfer the data copy of this image to the destination image:
	GameX.AccessPixels(dest);
	for(int y = dy1 ; y < dy2 ; y++)
	{
		if(dest->z_ymult != 1.0f && y != dest->z_fy1 && (int)(dest->z_ymult * (float)(y-1)) == (int)(dest->z_ymult * (float)(y))) continue;
		for(int x = dx1 ; x < dx2 ; x++)
		{
			if(dest->z_xmult != 1.0f && x != dest->z_fx1 && (int)(dest->z_xmult * (float)(x-1)) == (int)(dest->z_xmult * (float)(x))) continue;
			int offset = 4*(x-dx1+dxWidth*(y-dy1));
			GameX.DrawPixel(x,y,(int)data[0+offset],(int)data[1+offset],(int)data[2+offset],(int)data[3+offset]);
		}
	}
	GameX.EndPixelAccess();

	// finally, deallocate the temporary copy memory
	SAFE_DELETE_ARRAY(data);
}

void ImageX::CopyToAlpha (ImageX * dest) const
{
	if(dest == NULL || dest == VIEWPORT || z_status != IMG_OK) return;

	if(dest->z_status != IMG_OK) // if copying to image that hasn't been loaded or created yet,
		dest->z_CreateWithUse(z_xres,z_yres,z_usage|LOAD_ALPHA); // create it before copying into it

	if(!(dest->z_usage & LOAD_ALPHA)) { // if destination has no alpha channel, convert it to have one:
		ImageUsage use = dest->z_usage|LOAD_ALPHA;
		if(use & LOAD_MASKED) use ^= LOAD_MASKED;
		dest->ConvertUsageTo(use);
	}

	int rmin = GameX.z_win_maskminr;
	int gmin = GameX.z_win_maskming;
	int bmin = GameX.z_win_maskminb;

	int sy1 = z_fy1;
	int sy2 = z_fy2;
	int sx1 = z_fx1;
	int sx2 = z_fx2;
	int dy1 = dest->z_fy1;
	int dy2 = dest->z_fy2;
	int dx1 = dest->z_fx1;
	int dx2 = dest->z_fx2;
///	int x2 = (z_fx2 >= dest->z_xres) ? dest->z_surf_width : z_fx2;

	int sxWidth = sx2-sx1;
	int syHeight = sy2-sy1;
	
	// allocate enough data to hold source pixels matching (in size) the destination:
	int dxWidth = dx2-dx1;
	int dyHeight = dy2-dy1;
	int size = (dxWidth)*(dyHeight);
	unsigned char * data = new unsigned char[size];

	// make a temporary grayscale copy of this image, scaled to destination size if necessary:
	GameX.AccessPixels(this);
	int c[4];
	if(sxWidth == dxWidth && syHeight == dyHeight) { // if no scaling needed:
		for(int y = dy1 ; y < dy2 ; y++)
		for(int x = dx1 ; x < dx2 ; x++)
		{
			GameX.ReadPixel(x,y,c[0],c[1],c[2],c[3]);
			data[(x-dx1+(dxWidth)*(y-dy1))] = (unsigned char)((int)(c[0]*RED_LUMINANCE + c[1]*GREEN_LUMINANCE + c[2]*BLUE_LUMINANCE) /** c[3] / 255*/);
		}
	} else { // otherwise scale with nearest-neighbor method:
		double xmult = (double)sxWidth/(double)dxWidth;
		double ymult = (double)syHeight/(double)dyHeight;
		for(int y = dy1 ; y < dy2 ; y++)
		{
			int gy = (int)((double)y*ymult);
			if(z_ymult != 1.0f && gy != z_fy1 && (int)(z_ymult * (float)(gy-1)) == (int)(z_ymult * (float)(gy))) continue;
			for(int x = dx1 ; x < dx2 ; x++)
			{
				int gx = (int)((double)x*xmult);
				if(z_xmult != 1.0f && gx != z_fx1 && (int)(z_xmult * (float)(gx-1)) == (int)(z_xmult * (float)(gx))) continue;
				GameX.ReadPixel(gx,gy,c[0],c[1],c[2],c[3]);
				data[(x-dx1+(dxWidth)*(y-dy1))] = (unsigned char)((int)(c[0]*RED_LUMINANCE + c[1]*GREEN_LUMINANCE + c[2]*BLUE_LUMINANCE) /** c[3] / 255*/);
			}
		}
	}
	GameX.EndPixelAccess();

	// now transfer the data copy of this image to the destination image:
	GameX.AccessPixels(dest);
	for(int y = dy1 ; y < dy2 ; y++)
	{
		if(dest->z_ymult != 1.0f && y != dest->z_fy1 && (int)(dest->z_ymult * (float)(y-1)) == (int)(dest->z_ymult * (float)(y))) continue;
		for(int x = dx1 ; x < dx2 ; x++)
		{
			if(dest->z_xmult != 1.0f && x != dest->z_fx1 && (int)(dest->z_xmult * (float)(x-1)) == (int)(dest->z_xmult * (float)(x))) continue;
			int offset = (x-dx1+dxWidth*(y-dy1));
			GameX.ReadPixel(x,y,c[0],c[1],c[2]);
			c[3] = (int)(data[offset]);
			if(dest->z_usage & LOAD_MASKED)
			{
				if(c[3] >= gmin && c[0]<rmin && c[1]<gmin && c[2]<bmin)
					{c[0] = rmin; c[1] = gmin; c[2] = bmin; c[3] = 255;}
				else if(c[3]<gmin && (c[0] >= rmin || c[1] >= gmin || c[2] >= bmin))
					{c[0] = 0; c[1] = 0; c[2] = 0; c[3] = 0;}
			}
			GameX.DrawPixel(x,y,c[0],c[1],c[2],c[3]);
		}
	}
	GameX.EndPixelAccess();

	// finally, deallocate the temporary copy memory
	SAFE_DELETE_ARRAY(data);
}

bool ImageX::CopyToClipboard ()
{
	ImageExt ie;
	return ie.CopyToClipboard(this);
}


void ImageX::UseWith (ImageUsage use)
{
	z_usage = use;
	if(!(z_options & IMG_MERGEALPHA))
	{
		switch(GameX.GetBPP())
		{
		case 8:
			z_ConvertFormat(IMG_INDEXED);
			break;
		case 16:
			z_ConvertFormat(IMG_HIGHCOLOR);
			break;
		case 24:
			z_ConvertFormat(IMG_TRUECOLOR);
			break;
		case 32:
			z_ConvertFormat(IMG_TRUECOLORX);
			break;
		}
	}
	z_size = z_xres*z_yres;
	z_xmult = 1.0f;
	z_ymult = 1.0f;

	bool makeSurface = true;
	bool saveData = !GameX.z_CanDeleteMData();

	if(z_options & IMG_ALPHA)
	{
		if(use & LOAD_MASKED)
			z_ConvertAlphaToMask();					   			
	}
	else
	{
		if(z_options & IMG_MERGEALPHA)
			z_options |= IMG_ALPHA;
		else
		if(use & LOAD_ALPHA)
			if(!(use & LOAD_MASKED))
				z_ConvertMaskToAlpha();
	}
///	if(!((z_usage & LOAD_ALPHA) && (z_options & IMG_ALPHA) && !(z_usage & LOAD_MASKED)))
	if(!(z_usage & LOAD_ALPHA) || !(z_options & IMG_ALPHA) || (z_usage & LOAD_MASKED))
	{
		if(z_options & IMG_ALPHA)
		{
			if(z_options & IMG_MERGEALPHA)
			{
				TurnOff(IMG_MERGEALPHA);
				if(GameX.GetBPP() == 8) z_ConvertFormat(IMG_INDEXED);
				if(GameX.GetBPP() == 16) z_ConvertFormat(IMG_HIGHCOLOR);
				if(GameX.GetBPP() == 24) z_ConvertFormat(IMG_TRUECOLOR);
			}
			TurnOff(IMG_ALPHA);
		}
		SAFE_DELETE_ARRAY(z_alpha);
	}
	if(makeSurface)
		z_AllocateSurface();

	if(z_hassurface)
	{
		z_DataToSurface();
		if(!saveData)
			DeleteData();
	}
	else
	{
		int adjusted_pitch = z_xres;
		SAFE_DELETE_ARRAY(z_ytable);
		z_ytable = new int[z_surf_height+32];
		for(int y = 0; y <= z_surf_height; y++)
			z_ytable[y] = y * adjusted_pitch;
	}
	
	z_status = IMG_OK;

	ResetFilterRect();
}

template<typename type>
void MaskToAlphaHelper(type* pData, unsigned char* pAlphaData, int size)
{
	for(int n = 0; n<size; n++)
	{
		if(*pData == 0)
			*pAlphaData = 0;
		else
			*pAlphaData = 255;
		*pAlphaData++;
		*pData++;
	}
}

// FIXME: this is wrong in 24 bit color mode
void ImageX::z_ConvertMaskToAlpha ()
{
	// main reason for doing this:
	// D3D renders color key edges in a very ugly way,
	// but it handles alpha channels just fine (esp. with rotation).

	if(GetAlpha(z_options) != IMG_ALPHA)
	{
		SAFE_DELETE_ARRAY(z_alpha);
		z_alpha = new unsigned char[z_xres * z_yres];
		if(z_options & IMG_TRUECOLORX)
			MaskToAlphaHelper((unsigned long*)z_data, z_alpha, z_size);
		else
			MaskToAlphaHelper((unsigned short*)z_data, z_alpha, z_size);
		z_options |= IMG_ALPHA;
	}
}

void ImageX::SetPixel (int x, int y, int r, int g, int b, int a)
{
	if(x < 0 || y < 0 || x >= min(z_xres+1,z_surf_width) || y >= min(z_yres+1,z_surf_height))
		return;

	if(r < 0) r = 0;
	if(g < 0) g = 0;
	if(b < 0) b = 0;
	if(a < 0) a = 0;
	if(r > 255) r = 255;
	if(g > 255) g = 255;
	if(b > 255) b = 255;
	if(a > 255) a = 255;

	GameX.AccessPixels(this);
	GameX.DrawPixel(x,y,r,g,b,a);
	GameX.EndPixelAccess();
}

void ImageX::GetPixel (int x, int y, int &r, int &g, int &b, int &a)
{
	if(x < 0 || y < 0 || x >= z_xres || y >= z_yres)
	{
		r = 0; g = 0; b = 0; a = 0;
		return;
	}
	GameX.AccessPixels(this);
	GameX.ReadPixel(x,y,r,g,b,a);
	GameX.EndPixelAccess();
}

void ImageX::z_ConvertFormat (ImageOps ops)
{	
	unsigned char * newdata;
	
	switch (GetFormat(z_options)) {			// Source Format:
		case IMG_INDEXED:					//   Indexed Color (8 bit)
			switch (GetFormat(ops))	{	// Target Format:
				case IMG_HIGHCOLOR:	//   High Color (16 bit)
					assertMsg(0, "Unsupported color conversion operation (8 to 16 bit).") ;
					break;
				case IMG_TRUECOLOR:	//   True Color (24 bit)
					assertMsg(0, "Unsupported color conversion operation (8 to 24 bit).") ;
					break;
				case IMG_TRUECOLORX:	//   True Color with X channel (32 bit)
					assertMsg(0, "Unsupported color conversion operation (8 to 32 bit).") ;
					break;
			}
			break;
		case IMG_HIGHCOLOR:					//   High Color (16 bit)
			switch (GetFormat(ops))	{	// Target Format:
				case IMG_INDEXED:	//   High Color (8 bit)
					GameX.z_SafeMessageBox("Error: Unsupported color conversion operation (16 to 8 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break; // not yet supported
				case IMG_TRUECOLOR:	{ //   True Color (24 bit)
					unsigned short *pDataOld;
					unsigned char *pDataNew;
					newdata = new unsigned char[z_size * GetBytesPerPixel(ops)];			
					pDataOld = (unsigned short *) z_data; 
					pDataNew = (unsigned char *) newdata;

					if(!GameX.z_win_shiftr)
						GameX.z_ReportProblem(GAMEX_NO_INIT) ;

					for(int n = 0; n<z_size; n++)
					{
						int clr = *pDataOld++;
						int blue  = (((clr & 0x001F) >> 0 ) << 3);
						int green = (((clr & 0x07E0) >> 5 ) << 2);
						int red   = (((clr & 0xF800) >> 11) << 3);
						*pDataNew++ = red;
						*pDataNew++ = green;
						*pDataNew++ = blue;
					}
					SAFE_DELETE_ARRAY(z_data);
					z_data = newdata; newdata = NULL;		
					break;
				}
				case IMG_TRUECOLORX: {	//   True Color with X channel (32 bit)
					unsigned short *pDataOld;
					unsigned long *pDataNew;
					newdata = new unsigned char[z_size * GetBytesPerPixel(ops)];			
					pDataOld = (unsigned short *) z_data; 
					pDataNew = (unsigned long *) newdata;

					if(!GameX.z_win_shiftr)
						GameX.z_ReportProblem(GAMEX_NO_INIT) ;

					for(int n = 0; n<z_size; n++)
					{
						int clr = *pDataOld++;
						int blue  = (((clr & 0x001F) >> 0 ) << 3);
						int green = (((clr & 0x07E0) >> 5 ) << 2);
						int red   = (((clr & 0xF800) >> 11) << 3);
						*pDataNew++ = (red << GameX.z_win_shiftr) + (green << GameX.z_win_shiftg) + (blue /*<< GameX.z_win_shiftb*/);
					}
					SAFE_DELETE_ARRAY(z_data);
					z_data = newdata; newdata = NULL;
					break;
				}
			}
			break;							// Source Format:
		case IMG_TRUECOLOR:					//   True Color (24 bit)
			switch (GetFormat(ops))	{			// Target Format:
				case IMG_INDEXED:			//   Indexed Color (8 bit)
					GameX.z_SafeMessageBox("Error: Unsupported color conversion operation (24 to 8 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break;
				case IMG_HIGHCOLOR:	{			//   High Color (16 bit)
					unsigned char *pDataOld;
					unsigned short *pDataNew;
					newdata = new unsigned char[z_size * GetBytesPerPixel(ops)];			
					pDataOld = (unsigned char *) z_data; 
					pDataNew = (unsigned short *) newdata;

					if(!GameX.z_win_shiftr)
						GameX.z_ReportProblem(GAMEX_NO_INIT) ;

					for(int n = 0; n<z_size; n++)
					{
						*pDataNew = (*pDataOld++ >> 3) << GameX.z_win_shiftr;
						*pDataNew |= (*pDataOld++ >> 2) << GameX.z_win_shiftg;
						*pDataNew++ |= (*pDataOld++ >> 3) /*<< GameX.z_win_shiftb*/;
					}
					SAFE_DELETE_ARRAY(z_data);
					z_data = newdata; newdata = NULL;
					break;
				}
				case IMG_TRUECOLORX: {	//   True Color with X channel (32 bit)
					unsigned char *pDataOld;
					unsigned long *pDataNew;
					newdata = new unsigned char[z_size * GetBytesPerPixel(ops)];			
					pDataOld = (unsigned char *) z_data; 
					pDataNew = (unsigned long *) newdata;

					if(!GameX.z_win_shiftr)
						GameX.z_ReportProblem(GAMEX_NO_INIT) ;

					for(int n = 0; n<z_size; n++)
					{
						int newPixel = 0;
						newPixel |= *pDataOld++ << GameX.z_win_shiftr;
						newPixel |= *pDataOld++ << GameX.z_win_shiftg;
						newPixel |= *pDataOld++ /*<< GameX.z_win_shiftb*/;
						*pDataNew++ = newPixel;
					}
					SAFE_DELETE_ARRAY(z_data);
					z_data = newdata; newdata = NULL;
					break;
				}
			}
			break;							// Source Format:
		case IMG_TRUECOLORX:				//   True Color with X channel (32 bit)
			switch (GetFormat(ops))	{		// Target Format:
				case IMG_INDEXED:			//   Indexed Color (8 bit)
					GameX.z_SafeMessageBox("Error: Unsupported color conversion operation (32 to 8 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break;
				case IMG_HIGHCOLOR: {			//   High Color (16 bit)
					unsigned long *pDataOld;
					unsigned short *pDataNew;
					newdata = new unsigned char[z_size * GetBytesPerPixel(ops)];			
					pDataOld = (unsigned long *) z_data; 
					pDataNew = (unsigned short *) newdata;

					if(!GameX.z_win_shiftr)
						GameX.z_ReportProblem(GAMEX_NO_INIT) ;

					for(int n = 0; n<z_size; n++)
					{
						int oldPixel = *pDataOld++;
						int red = ((oldPixel & 0x00FF0000) >> 16);
						int green = ((oldPixel & 0x0000FF00) >> 8);
						int blue = ((oldPixel & 0x000000FF) >> 0);
						int newPixel = ((red >> 3) << 11) | ((green >> 2) << 5) | ((blue >> 3) << 0);
						*pDataNew++ = newPixel;
					}
					SAFE_DELETE_ARRAY(z_data);
					z_data = newdata; newdata = NULL;
					break;
				}
				case IMG_TRUECOLOR: {			//   True Color (24 bit)
					unsigned long *pDataOld;
					unsigned char *pDataNew;
					newdata = new unsigned char[z_size * GetBytesPerPixel(ops)];			
					pDataOld = (unsigned long *) z_data; 
					pDataNew = (unsigned char *) newdata;

					if(!GameX.z_win_shiftr)
						GameX.z_ReportProblem(GAMEX_NO_INIT) ;

					for(int n = 0; n<z_size; n++)
					{
						int oldPixel = *pDataOld++;
						int red = ((oldPixel & 0x00FF0000) >> 16);
						int green = ((oldPixel & 0x0000FF00) >> 8);
						int blue = ((oldPixel & 0x000000FF) >> 0);
						*pDataNew++ = red;
						*pDataNew++ = green;
						*pDataNew++ = blue;
					}
					SAFE_DELETE_ARRAY(z_data);
					z_data = newdata; newdata = NULL;
					break;
				}
			}
			break;
	}
	TurnOff (GetFormat(z_options));
	TurnOn (ops);
}

template<typename type>
void AlphaToMaskHelper(type* pData, unsigned char* pAlphaData, int size)
{
	for(int n = 0; n<size; n++)
	{
		if(*pData == 0) *pData = 1;
		if(*pAlphaData < 16) *pData = 0;
		*pData++;
		*pAlphaData++;
	}
}

// FIXME: this is wrong in 24 bit color mode and untested in 32
void ImageX::z_ConvertAlphaToMask ()
{
	if(GetAlpha(z_options) == IMG_ALPHA)
	{
		switch (GetFormat(z_options))
		{
		case IMG_HIGHCOLOR:
			AlphaToMaskHelper((unsigned short*)z_data, z_alpha, z_size);
			break;
		case IMG_TRUECOLOR:
		case IMG_TRUECOLORX:
			AlphaToMaskHelper((unsigned long*)z_data, z_alpha, z_size);
			break;
		}

		TurnOff (IMG_ALPHA);
		TurnOn (IMG_MASKED);
	}
}

// Image rotation function for backward compatibility
// (DrawImage or DrawImageEx should be used for rotation now)
void ImageX::Rotate (ImageX *dest, float ang, float scale)
{
	// Fill the destination with pure black, like Rotate used to do.
	GameX.ResetDrawStates();
	GameX.SetDrawDestination(dest);
	GameX.FillScreen(ColorX(0,0,0,0));

	// Now draw the rotated image into the center of the destination.
	// Note that this requires destination to be loaded with LOAD_TARGETABLE.
	GameX.SetDrawDestination(dest);
	GameX.SetDrawMode(DRAW_PLAIN|DRAWOP_CENTERED);
	GameX.DrawImage(this, dest->z_xres*0.5f, dest->z_yres*0.5f, ang,scale);
	GameX.ResetDrawStates();
}

/*
		
void ImageX::RotateImage (ImageX *dest, float ang, float scale, bool toVideo)
{
	// NOTE: ASSUMES BOTH IMAGES ARE IMG_HIGHCOLOR

	if(dest != NULL && z_data != NULL)
	{
		unsigned short *src_pix;
		unsigned short *dest_pix;
		unsigned short *dest_start;
		float delt_x, delt_y;
		float start_x, start_y;
		float off_x, off_y;
		float pix_x, pix_y, h;
		int y, cnt;
		int rowSize;

		h = (float) sqrt(dest->z_xres * dest->z_xres + dest->z_yres * dest->z_yres);
		start_x = (z_xres/2.0f) + cosf((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		start_y = (z_yres/2.0f) - sinf((135-ang)*DEGtoRAD)*(h/2.0f)*(z_xres/z_yres)/scale+1; // the +1 fixes a vertical off-by-one bug
		delt_x = cosf(ang*DEGtoRAD)/scale;
		delt_y = sinf(ang*DEGtoRAD)*(z_xres/z_yres)/scale;
		off_x = delt_x * dest->z_xres;
		off_y = delt_y * dest->z_xres;
		if(toVideo)
		{
			DDSURFACEDESC2 ddsd;	
			ZeroMemory (&ddsd, sizeof (ddsd)) ;
			ddsd.dwSize = sizeof (ddsd) ;

 			HRESULT z_status = dest->z_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);	// lock drawing surface
			if(z_status != DD_OK)
			{
				GameX.z_DirectDrawError(z_status);
				return;
			}
			dest_start = (unsigned short *) ddsd.lpSurface;
			rowSize = ddsd.lPitch >> 1; // >> 1 for 16 bits
		}
		else
		{
			dest_start = (unsigned short*) dest->z_data;
			rowSize = dest->z_xres;
		}
		for(y = 0; y<dest->z_yres; y++) {	
			cnt = dest->z_xres;						// Start with full pixel set in line
			pix_x = start_x + off_x-1; //-1 fixes off-by-1 bug				// Goto end of line and scan backwards
			pix_y = start_y + off_y;
			dest_pix = dest_start + cnt-1;
			while((pix_x<0 || pix_x >= z_xres || pix_y<0 || pix_y >= z_yres) && cnt>0)
			{
				pix_x -= delt_x;					// Scan until end of span found
				pix_y -= delt_y;
				cnt--;
				*dest_pix-- = 0;
			}	
			dest_pix = dest_start;
			pix_x = start_x;						// Goto beginning of line and scan forward
			pix_y = start_y;
			while((pix_x<0 || pix_x >= z_xres || pix_y<0 || pix_y >= z_yres) && cnt>0)
			{
				pix_x += delt_x;					// Scan until start of span found
				pix_y += delt_y;
				cnt--;
				*dest_pix++ = 0;
			}
			while(cnt>0)
			{
				src_pix = ((unsigned short*) z_data) + (int) pix_y*z_xres + (int) pix_x;
				*dest_pix++ = *src_pix;			
				pix_x += delt_x; pix_y += delt_y;
				cnt--;
			}
			dest_start += rowSize;
			start_x -= delt_y;
			start_y += delt_x;
		}
		if(toVideo)
{
 			z_status = dest->z_surface->Unlock(NULL); 	// unlock drawing surface
			if(z_status != DD_OK){
				GameX.z_DirectDrawError(z_status);
				return;
			}
		}
	}
}

void ImageX::RotateAlpha (ImageX *dest, float ang, float scale)
{
	if(dest->z_options & IMG_ALPHA && z_alpha != NULL && dest->z_alpha != NULL)
	{
		unsigned char *src_pix;
		unsigned char *dest_pix;
		unsigned char *dest_start;
		float delt_x, delt_y;
		float start_x, start_y;
		float off_x, off_y;
		float pix_x, pix_y, h;
		int y, cnt;

		h = (float) sqrt(dest->z_xres * dest->z_xres + dest->z_yres * dest->z_yres);
		start_x = (z_xres/2.0f) + cosf((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		start_y = (z_yres/2.0f) - sinf((135-ang)*DEGtoRAD)*(h/2.0f)/scale+1; // the +1 is to work around an off-by-one bug
		delt_x = cosf(ang*DEGtoRAD)/scale;
		delt_y = sinf(ang*DEGtoRAD)/scale;
		off_x = delt_x * dest->z_xres;
		off_y = delt_y * dest->z_xres;	
		dest_start = (unsigned char*) dest->z_alpha;
		for(y = 0; y<dest->z_yres; y++) {	
			cnt = dest->z_xres;						// Start with full pixel set in line
			pix_x = start_x + off_x-1; // -1 avoids off-by-one bug		// Goto end of line and scan backwards
			pix_y = start_y + off_y;
			dest_pix = dest_start + cnt-1;
			while((pix_x<0 || pix_x >= z_xres || pix_y<0 || pix_y >= z_yres) && cnt>0)
			{
				pix_x -= delt_x;					// Scan until end of span found
				pix_y -= delt_y;
				cnt--;
				*dest_pix-- = 255;
			}	
			dest_pix = dest_start;
			pix_x = start_x;						// Goto beginning of line and scan forward
			pix_y = start_y;
			while((pix_x<0 || pix_x >= z_xres || pix_y<0 || pix_y >= z_yres) && cnt>0)
			{
				pix_x += delt_x;					// Scan until start of span found
				pix_y += delt_y;
				cnt--;
				*dest_pix++ = 255;
			}
			while(cnt>0)
			{
				src_pix = ((unsigned char*) z_alpha) + (int) pix_y*z_xres + (int) pix_x;
				*dest_pix++ = *src_pix;			
				pix_x += delt_x; pix_y += delt_y;
				cnt--;
			}
			dest_start += dest->z_xres;
			start_x -= delt_y;
			start_y += delt_x;
		}
	}
}
*/

ImageExt::ImageExt ()
{
	m_pImg = NULL;
}

ImageExt::ImageExt (ImageX* pImg)
{
	m_pImg = pImg;
}

ImageExt::~ImageExt ()
{
	m_pImg = NULL;
}

bool ImageExt::Load (const char * filename, ImageX * img, bool alpha)
{
	ImageUsage use = (alpha) ? (LOAD_ALPHA) : 0;
	return LoadWithUse(filename,img,use);
}

bool ImageExt::LoadWithUse (const char * filename, ImageX * img, ImageUsage use)
{
	char tempFilename [MAX_PATH];
	strcpy(tempFilename, filename);

	img->Destroy();
	img->Reset();

	strcpy(img->z_filename,tempFilename);

	if(GameX.z_win_log_debug)
	{
		if(GameX.z_GetStatus() == GAMEX_READY)
		{
			const char* filenamePtr = REMOVE_TRAILING_SEPARATOR(filename);

			int z;
			char debug_message [MAX_PATH];
			for(z = 0 ; filenamePtr[z] ; z++)
			{
				debug_message[z] = filenamePtr[z];
			}
			debug_message[z++] = ':';
			while(z < 24)
			{
				debug_message[z++] = ' ';
			}
			debug_message[z++] = (char)0;
			debug.Output(debug_message); 
		}
	}

	bool success = false;


	const char * extension = strrchr(filename, '.'); // string of text from the last dot onward

	if(!extension || (_strnicmp(extension, ".TIF", 4) == 0))
		success = LoadTiff(filename, img);

	if(!success && (!extension || (_stricmp(extension, ".BMP") == 0 || _stricmp(extension, ".DIB") == 0)))
		success = LoadBMP(filename, img);

	if(!success && (!extension || (_stricmp(extension, ".PNG") == 0)))
		success = LoadPNG(filename, img);

	if(!success)
		success = LoadJPGorGif(filename, img);

	if(!success)
	{
		// something went wrong, either the file wasn't found
		// or it's in a nonstandard format such as [A4 R4 G4 B4] or [R5 G6 B5]

		assertMsg2(0, "The image in file %s did not load.\nMake sure the image is in that location and that this game's executable is in the correct location to find it.", filename);

		img->Create(64,64,true); // arbitrary size
		if(debug.IsRunningDebug())
		{
			// debugging - make it look like puke
			img->Fill(255,192,0,255);
			img->AddNoise(-128, -128, 128, -128);
		}
		else
		{
			// not debugging - make it black/invisible
			img->Fill(0,0,0,0); 
		}

		return false;
	}

	img->UseWith(use);

	if(success) img->z_non_empty = true;

	return success;
}

bool ImageExt::Save (const char * filename, ImageX * img)
{
	bool ok = false;
	const char * extension = strrchr(filename, '.'); // string of text from the last dot onward
	if(extension && (_stricmp(extension, ".TIF") == 0 || _stricmp(extension, ".TIFF") == 0)) { // case insensitive match
		ok = SaveTiff(filename, img);
	} else { // since the only other format we can save to is BMP, assume that and warn if it isn't
		LPDIRECTDRAWSURFACE7 surface;
		if(img == VIEWPORT)
		{
			surface = GameX.z_GetBackSurf();
		}
		else
		{
			if(img->z_status != IMG_OK)
				return false;
			surface = img->z_surface;
		}

		HDC hdc, hmemdc;
		HBITMAP hbitmap, hprevbitmap;
		DDSURFACEDESC2 ddsd;
		surface->GetDC(&hdc);
		hmemdc = CreateCompatibleDC(hdc); 
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		surface->GetSurfaceDesc(&ddsd);
		hbitmap = CreateCompatibleBitmap(hdc, ddsd.dwWidth, ddsd.dwHeight);
		hprevbitmap = (HBITMAP) SelectObject(hmemdc, hbitmap);
		BitBlt(hmemdc, 0, 0, ddsd.dwWidth, ddsd.dwHeight, hdc, 0, 0, SRCCOPY);
			
		ok = SaveBmp(filename,hbitmap);

		SelectObject(hmemdc,hprevbitmap);
		DeleteDC(hmemdc);
		surface->ReleaseDC(hdc);

		if(!(extension && (_stricmp(extension, ".BMP") == 0 || _stricmp(extension, ".DIB") == 0)))
		{
			char errstr [256], fixstr [64];
			if(extension == NULL) sprintf(fixstr, "rename it to \"%s.bmp\".", filename);
			else				  sprintf(fixstr, "replace \"%s\" with \".bmp\".", extension);
			sprintf(errstr,"GameX has been requested to save an ImageX to \"%s\", but GameX only supports saving to .BMP or .TIFF files.\nTherefore, to open \"%s\", you will need to %s",filename,filename,fixstr);
			GameX.z_SafeMessageBox(errstr, "GameX Warning", MB_OK|MB_ICONWARNING);
		}
	}

	return ok;
}

bool ImageExt::CopyToClipboard(const ImageX * img)
{
	return GameX.CopyToClipboard(img);
}

void ImageX::ConvertRGBToLab(int r, int g, int b, int & L, int & A, int & B)
{
	float x = (0.412453f*r + 0.357580f*g + 0.180423f*b) * (1.0f / (255.0f * 0.950456f));
	float y = (0.212671f*r + 0.715160f*g + 0.072169f*b) * (1.0f / 255.0f);
	float z = (0.019334f*r + 0.119193f*g + 0.950227f*b) * (1.0f / (255.0f * 1.088754f));

	float fx, fy, fz;

	if(y > 0.008856f)
	{
		fy = powf(y, 0.333333333f);
		L = Round(116.0f*fy + (-16.0f));
	} else
	{
		fy = 7.787f*y + (16.0f/116.0f);
		L = Round(903.3f*y);
	}

	if(x > 0.008856f)
		fx = powf(x, 0.333333333f);
	else
		fx = 7.787f*x + (16.0f/116.0f);

	A = Round(500.0f*(fx - fy));

	if(z > 0.008856f)
		fz = powf(z, 0.333333333f);
	else
		fz = 7.787f*z + (16.0f/116.0f);

	B = Round(200.0f*(fy - fz));
}

void ImageX::ConvertLabToRGB(int L, int A, int B, int &r, int &g, int &b)
{
	float y;
	float fy = (L + 16.0f) * (1.0f / 116.0f);
	fy = fy*fy*fy;
	if(fy < 0.008856f)
	{
		fy = L * (1.0f/903.3f);
		y = fy;
		fy = 7.787f*fy + (16.0f/116.0f);
	}
	else
	{
		y = fy;
		fy = powf(fy, 0.333333333f);
	}

	float x, fx = A * (1.0f/500.0f) + fy;
	float z, fz = fy - B * (1.0f/200.0f);

	if(fx > 0.206893f)
		x = fx*fx*fx;
	else
		x = (fx - (16.0f/116.0f)) * (1.0f / 7.787f);

	if(fz > 0.206893f)
		z = fz*fz*fz;
	else
		z = (fz - (16.0f/116.0f)) * (1.0f / 7.787f);

	x *= (0.950456f * 255.0f);
	y *= (255.0f);
	z *= (1.088754f * 255.0f);

	r = Round(3.240479f*x - 1.537150f*y - 0.498535f*z);
	g = Round(-0.969256f*x + 1.875992f*y + 0.041556f*z);
	b = Round(0.055648f*x - 0.204043f*y + 1.057311f*z);

	if(r<0) r = 0; if(r>255) r = 255;
	if(g<0) g = 0; if(g>255) g = 255;
	if(b<0) b = 0; if(b>255) b = 255;
}


void ColorX::InvertBrightness ()
{
	for(int i = 0 ; i < 4 ; i++)
	{
		int L, A, B;
		ImageX::ConvertRGBToLab(c[i][0], c[i][1], c[i][2], L, A, B);
		ImageX::ConvertLabToRGB(150-L, A, B, c[i][0], c[i][1], c[i][2]);
	}
}

void ColorX::InvertHue ()
{
	for(int i = 0 ; i < 4 ; i++)
	{
		int L, A, B;
		ImageX::ConvertRGBToLab(c[i][0], c[i][1], c[i][2], L, A, B);
		ImageX::ConvertLabToRGB(L, -A, -B, c[i][0], c[i][1], c[i][2]);
	}
}

void ColorX::ConvertToGrayscale()
{
	for(int i = 0 ; i < 4 ; i++)
	{
		const int gray = c[i][0]*((int)(RED_LUMINANCE*255))/255 + c[i][1]*((int)(GREEN_LUMINANCE*255))/255 + c[i][2]*((int)(BLUE_LUMINANCE*255))/255;

		for(int j = 0; j < 3 ; j++)
			c[i][j] = gray;
	}
}

